// ***************************************************************************
// CPairedEndSort - resolves paired-end reads and creates an alignment archive
//                  sorted by reference sequence position.
// ---------------------------------------------------------------------------
// (c) 2006 - 2009 Michael Strmberg
// Marth Lab, Department of Biology, Boston College
// ---------------------------------------------------------------------------
// Dual licenced under the GNU General Public License 2.0+ license or as
// a commercial license with the Marth Lab.
// ***************************************************************************

#pragma once

#include <iostream>
#include <list>
#include <set>
#include <map>
#include <string>
#include "sqlite3.h"
#include "AlignmentReader.h"
#include "AlignmentStatus.h"
#include "AlignmentWriter.h"
#include "AlignedRead.h"
#include "ConsoleUtilities.h"
#include "ReadGroup.h"
#include "Mosaik.h"
#include "ProgressBar.h"
#include "ProgressCounter.h"
#include "SequencingTechnologies.h"
#include "UnorderedMap.h"

using namespace std;

// here we assume that we'll need space for Sanger length reads (reference bases,
// query bases, query base qualities)
static double DEFAULT_CONFIDENCE_INTERVAL = 0.9973;
#define GAP '-'

// define our serialization status flags
#define PE_IS_LONG_READ           1
#define PE_IS_REVERSE_STRAND      2
#define PE_IS_MATE_REVERSE_STRAND 4
#define PE_IS_FIRST_MATE          8
#define PE_IS_RESOLVED_AS_PAIR    16
#define PE_WAS_RESCUED            32

#define SQL_BUFFER_SIZE 10240

// define our model count data structure
struct ModelType {
	unsigned char ID;
	unsigned int Count;

	// constructor
	ModelType(void)
		: ID(0)
		, Count(0)
	{}

	// our less-than operator
	bool operator<(const ModelType& mt) const {
		return mt.Count < Count;
	}
};

class CPairedEndSort {
public:
	// constructor
	CPairedEndSort(const unsigned int numCachedReads);
	// destructor
	~CPairedEndSort(void);
	// configures which read pair types should be resolved
	void ConfigureResolution(const bool uo, const bool uu, const bool um, const bool mm);
	// allows any fragment length when evaluating unique mate-pairs
	void EnableAllUniqueFragmentLengths(void);
	// enables consed renaming
	void EnableConsedRenaming(void);
	// enables duplicate read filtering
	void EnableDuplicateFiltering(const string& duplicateDirectory);
	// enables the sampling of all read pairs
	void EnableFullFragmentLengthSampling(void);
	// resolves the paired-end reads found in the specified input file
	void ResolvePairedEndReads(const string& inputFilename, const string& outputFilename);
	// sets the desired confidence interval
	void SetConfidenceInterval(const double& percent);

private:
	// define our sort configuration structure
	struct SortSettings {
		string DuplicateDirectory;
		string UnresolvedFilename;
		double ConfidenceInterval;
		unsigned int NumCachedReads;

		SortSettings() 
			: ConfidenceInterval(DEFAULT_CONFIDENCE_INTERVAL)
			, NumCachedReads(0)
		{}
	} mSettings;
	// define our boolean flags structure
	struct FlagData {
		bool AllowAllUniqueFragmentLengths;
		bool RemoveDuplicates;
		bool RenameMates;
		bool ResolveMM;
		bool ResolveUM;
		bool ResolveUO;
		bool ResolveUU;
		bool SampleAllFragmentLengths;

		FlagData()
			: AllowAllUniqueFragmentLengths(false)
			, RemoveDuplicates(false)
			, RenameMates(false)
			, ResolveMM(false)
			, ResolveUM(false)
			, ResolveUO(false)
			, ResolveUU(false)
			, SampleAllFragmentLengths(false)
		{}
	} mFlags;
	// retrieves an alignment from the specified temporary file and adds it to the specified list
	void AddAlignment(FILE* tempFile, const unsigned int owner, list<Alignment>& alignments);
	// retrieves an alignment from the specified temporary file
	bool GetAlignment(FILE* tempFile, const unsigned int owner, Alignment& al);
	// records the observed gaps in the specified reference 
	void RecordReferenceGaps(Alignment& al);
	// serializes the specified vector to a temporary file
	uint64_t Serialize(list<Alignment>& alignmentCache, const unsigned int numEntries);
	// our temporary file vector
	vector<string> mTempFiles;
	// our output buffer
	unsigned char* mBuffer;
	unsigned int mBufferLen;
	// our reference gap hash map vector and associated iterator
	vector<unordered_map<unsigned int, unsigned short> > mRefGapVector;
	unordered_map<unsigned int, unsigned short>::iterator mRefGapIter;
};
